import networkx
import pandas
import pyomo
import pyomo.opt
import pyomo.environ as pe
import scipy
import itertools
import logging

#==================================================================================
# With the following few lines of code, we are reading in the network specification.
#- We read in the set of nodes and associated node data from a nodes.csv file
#- We read in the set of arcs and their data from an arcs.csv file
#- We create a networkx object to store all of that.  This part is not strictly
# necessary -- you could go directly from the csv files to Pyomo LPs, but
# networkx is easier in terms of programming.  That ease of use and access is the
# only reason we make the networkx object.
#==================================================================================

def create_graph(nodefile, arcfile):
    g = networkx.DiGraph()
    # read the data files
    # We assume the first column of the node file is the node name
    # We also assume the arc file's first two columns are (node start, node end)
    nodes = pandas.read_csv(nodefile)
    arcs = pandas.read_csv(arcfile)
    # create the nodes and associated data
    nameCol = nodes.columns[0]
    for num,arow in nodes.iterrows():
        g.add_node( arow[nameCol], **(dict(arow))) 
    # create the arcs and associated data
    firstNodeCol = arcs.columns[0]
    secondNodeCol = arcs.columns[1]
    for num,arow in arcs.iterrows():
        g.add_edge( arow[firstNodeCol], arow[secondNodeCol], **(dict(arow))) 
    return g
    
# ============================================================================
# Lets define the primal LP.
# 
# This LP takes in attacks as constants, xbar, instead of variables x.
# ============================================================================

def create_pyomo_network_lp_primal(g,cost_field='Cost',capacity_field='Capacity',node_imbalance_field='SupplyDemand', interdiction_field='xbar'):
    # Create the model
    model = pe.ConcreteModel()
    # Tell pyomo to read in dual-variable information from the solver
    model.dual = pe.Suffix(direction=pe.Suffix.IMPORT)

    # Associate the graph with this model
    model.g = g

    # Create the problem data
    model.node_set = pe.Set( initialize=g.nodes() )
    model.edge_set = pe.Set( initialize=g.edges() )
    
    model.edge_cost = pe.Param( model.edge_set, 
            initialize=lambda model, i,j: model.g.edge[i][j].get(cost_field,0))
    model.edge_capacity = pe.Param( model.edge_set, 
            initialize=lambda model, i,j: model.g.edge[i][j].get(capacity_field,0))
    model.node_imbalance = pe.Param( model.node_set, 
            initialize=lambda model, i: model.g.node[i].get(node_imbalance_field,0))
    model.edge_delay_on = pe.Param( model.edge_set, 
            initialize=lambda model, i,j: model.g.edge[i][j].get(interdiction_field,0), mutable=True)

 # Compute nCmax
    nCmax = model.g.number_of_nodes() * max([data.get(cost_field, 0) for e0,e1,data in model.g.edges(data=True) ])
    model.nCmax = nCmax
    model.delay = pe.Param(initialize= 2*nCmax + 1, mutable=True)
    

    # Create the variables
    model.y = pe.Var(model.edge_set, domain=pe.NonNegativeReals)
    
    model.UnsatSupply = pe.Var(model.node_set, domain=pe.NonNegativeReals)
    model.UnsatDemand = pe.Var(model.node_set, domain=pe.NonNegativeReals)
    
    # Create the objective
    def obj_rule(model):
        return  pe.summation(model.edge_cost, model.y) + model.delay * pe.summation(model.edge_delay_on, model.y) + model.nCmax * pe.summation(model.UnsatSupply) + model.nCmax * pe.summation(model.UnsatDemand)

    model.OBJ = pe.Objective(rule=obj_rule, sense=pe.minimize)


    # Create the constraints, one for each node
    def flow_bal_rule(model, n):
        expr = 0
        if model.g.in_edges(n):
            expr += pe.summation(model.y, index=model.g.in_edges(n))
        if model.g.out_edges(n):
            expr += - pe.summation(model.y, index=model.g.out_edges(n))
        supply_node = (model.node_imbalance[n] < 0)
        demand_node = (model.node_imbalance[n] > 0)
        constr = (expr == model.node_imbalance[n] + model.UnsatSupply[n]*(supply_node) - model.UnsatDemand[n]*(demand_node) )
        if isinstance(constr, bool):
            if constr == True:
                return pe.Constraint.Feasible
            return pe.Constraint.Infeasible
        return constr

    model.FlowBalance = pe.Constraint(model.node_set, rule=flow_bal_rule)

    # Create the capacity constraints, one for each edge
    def capacity_rule(model, i, j):
        return model.y[(i,j)] <= model.edge_capacity[(i,j)]

    model.Capacity = pe.Constraint(model.edge_set, rule=capacity_rule)


    # Solve the model
    model.create()
    model= resolve_mip(model)

    # Print the model objective
    print 'Primal objective function value:', model.OBJ()
    return model

# ============================================================================
# This is a useful function, in case you want to re-solve a model in the
# interpreter.  Remember that if you dynamically want to change model
# parameters, you have to declare them as 'mutable' in the model.  If you look
# above, we've declared edge_delay_on and delay as mutable in this case.
# ============================================================================

def resolve_mip(model, tee=False):
    model.preprocess()
    solver = pyomo.opt.SolverFactory('cplex')
    results = solver.solve(model, tee=tee, keepfiles=False, options="mip_tolerances_integrality=1e-9 mip_tolerances_mipgap=0")

    # Check that we actually computed an optimal solution, load results
    if (results.solver.status != pyomo.opt.SolverStatus.ok):
        logging.warning('Check solver not ok?')
    if (results.solver.termination_condition != pyomo.opt.TerminationCondition.optimal):  
        logging.warning('Check solver optimality?')

    model.load(results)
    
    model.g.OBJ_val = model.OBJ()
    
    # Load solution data back into the networkx object 
    if hasattr(model, 'y'):
        for e in model.edge_set:
            model.g.edge[e[0]][e[1]]['flow_val'] = model.y[e].value

    if hasattr(model, 'dual'):
        for n in model.FlowBalance.iterkeys(): 
            model.g.node[n]['dual_val'] = model.dual[ model.FlowBalance[n] ]
    
    if hasattr(model, 'x'):
        for e in model.edge_set:
            model.g.edge[e[0]][e[1]]['attackActive'] = model.x[e].value
    
    if hasattr(model, 'UnsatDemand'):
        for n in model.node_set:
            model.g.node[n]['UnsatDemand'] = model.UnsatDemand[n].value
            model.g.node[n]['UnsatSupply'] = model.UnsatSupply[n].value
    
    return model


# ============================================================================
# Now, lets define the interdiction MIP
# ============================================================================

def create_pyomo_networkInterdiction_dualMIP(g,attacks=0,cost_field='Cost',capacity_field='Capacity',node_imbalance_field='SupplyDemand', interdiction_field='xbar'):
    # Create the model
    model = pe.ConcreteModel()

    # Associate the graph with this model
    model.g = g

    # Create the problem data
    model.node_set = pe.Set( initialize=g.nodes() )
    model.edge_set = pe.Set( within=model.node_set*model.node_set, initialize=g.edges() )
    
    model.edge_cost = pe.Param( model.edge_set, 
            initialize=lambda model, i,j: model.g.edge[i][j].get(cost_field,0))
    model.edge_capacity = pe.Param( model.edge_set, 
            initialize=lambda model, i,j: model.g.edge[i][j].get(capacity_field,0))
    model.node_imbalance = pe.Param( model.node_set, 
            initialize=lambda model, i: model.g.node[i].get(node_imbalance_field,0))
    model.edge_delay_on = pe.Param( model.edge_set, 
            initialize=lambda model, i,j: model.g.edge[i][j].get(interdiction_field,0), mutable=True)

    # Compute nCmax
    nCmax = model.g.number_of_nodes() * max([data.get(cost_field, 0) for e0,e1,data in model.g.edges(data=True) ])
    model.nCmax = nCmax
    model.delay = pe.Param(initialize= 2*nCmax + 1, mutable=True)
     
    model.attacks = pe.Param(initialize=attacks, mutable=True)

    # Create the variables
    model.rho = pe.Var(model.node_set, domain=pe.Reals)
    model.pi = pe.Var(model.edge_set, domain=pe.NonPositiveReals)
    
    model.x = pe.Var(model.edge_set, domain=pe.Binary)

    # Create the objective
    def obj_rule(model):
        return  pe.summation(model.edge_capacity, model.pi) + pe.summation(model.node_imbalance, model.rho)

    model.OBJ = pe.Objective(rule=obj_rule, sense=pe.maximize)

    # Set the x's for non-blockable arcs
    def edge_constraint_rule(model, i, j):
        attackable = (model.g.edge[i][j]['Attackable'] == 1)
        return model.rho[j] - model.rho[i] + model.pi[(i,j)] <=  model.edge_cost[(i,j)] + model.delay*model.x[(i,j)]*attackable

    model.DualEdgeConstraint = pe.Constraint(model.edge_set, rule=edge_constraint_rule)
    
    def unsat_constraint_rule(model, n):
        supply_node = (model.node_imbalance[n] < 0)
        demand_node = (model.node_imbalance[n] > 0)
        if (supply_node):
            return -model.rho[n] <= model.nCmax
        if (demand_node):
            return model.rho[n] <= model.nCmax
        return pe.Constraint.Feasible

    model.UnsatConstraint = pe.Constraint(model.node_set, rule=unsat_constraint_rule)
 
    # Create the interdiction budget constraint 
    def block_limit_rule(model):
        return pe.summation(model.x) <= model.attacks

    model.BlockLimit = pe.Constraint(rule=block_limit_rule)

    # Solve the model
    model.create()
    model= resolve_mip(model)

    # Print the model objective
    print 'Dual objective function value:', model.OBJ()
    return model

# ============================================================================
# Here, we are printing a nice report on the solution to the screen.  You have
# to pass a networkx object, with the flow_val field loaded.
# ============================================================================

def print_solution(g):

    edges = sorted(g.edges(data=True))
    for e0,e1,data in edges:
        if data.has_key('attackActive') and data['attackActive'] == 1:
            print 'Interdict arc %s -> %s'%(str(e0), str(e1))
    print
    
    nodes = sorted(g.nodes(data=True))
    for n,data in nodes:
        if data.has_key('UnsatSupply') and data['UnsatSupply'] > 0:
            print 'Remaining supply on node %s: %.2f'%(str(n), data['UnsatSupply'])
    for n,data in nodes:
        if data.has_key('UnsatDemand') and data['UnsatDemand'] > 0:
            print 'Remaining demand on node %s: %.2f'%(str(n), data['UnsatDemand'])
    print
    
    for e0,e1,data in edges:
        if data.has_key('flow_val') and data['flow_val'] > 0:
            print 'Flow on arc %s -> %s: %.2f'%(str(e0), str(e1), data['flow_val'])
    print

    print '----------'
    print 'Total cost = ', g.OBJ_val 


# ===============================================================
# Now that we've defined all our functions, lets do something....
#
# First, lets read the data. 
# ===============================================================

g = create_graph('sample_nodes_data.csv', 'sample_arcs_data.csv')

# ===============================================================
# Lets solve and print the flow when there are no
# interdictions.
# ===============================================================

mPrimal = create_pyomo_network_lp_primal(g, capacity_field='Capacity')

print 'Flow without interdictions:'
print_solution(mPrimal.g)

# ===============================================================
# Lets compute the optimal interdiction for 1 attack, recompute the
# flow response, and print the solution. 
# ===============================================================

print '---------------------------------'
print '---------------------------------'
# Creates and solves the dual
mDual = create_pyomo_networkInterdiction_dualMIP(g, attacks=1)
# We have to copy the X variables from the dual into xbar, and resolve the primal
# to compute the flow response
for e in mDual.edge_set:
    mDual.g.edge[e[0]][e[1]]['xbar'] = mDual.x[e].value
# Now re-solve the primal to get the flow
mPrimal = create_pyomo_network_lp_primal( mDual.g)

print 'Flow with %d attacks:\n'%(mDual.attacks.value)
print_solution(mDual.g)

